﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace DAMacTypesLib {
    public class RunningArray {
        //Array Stores Macro On/Off in bool[0], and Thread In/Out in bool[1]
        private Dictionary<IntPtr, List<bool[]>> hkRunning = new Dictionary<IntPtr, List<bool[]>>();
        private readonly object runningArrayLock = new object();
        private int hotkeyCount, macroOnCount;
        private NotifyIcon sysTray;
        private Icon SysNormal, SysRun;


        public RunningArray() : this(null, null, null) { }
        public RunningArray(NotifyIcon sysTray, Icon sysNormal, Icon sysRun) { this.sysTray = sysTray; this.SysNormal = sysNormal; this.SysRun = sysRun; }

        public void AddProcess(IntPtr process) {
            lock(runningArrayLock)
                if(!hkRunning.ContainsKey(process)) {
                    List<bool[]> runningList = new List<bool[]>(hotkeyCount);
                    for(int i = 0; i < hotkeyCount; i++) runningList.Add(new bool[2]);
                    hkRunning.Add(process, runningList);
                }
        }
        public void RemoveProcess(IntPtr process) {
            lock(runningArrayLock)
                if(hkRunning.ContainsKey(process)) {
                    foreach(bool[] i in hkRunning[process]) if(i[0] && --macroOnCount == 0 && sysTray != null) sysTray.Icon = SysNormal;
                    hkRunning.Remove(process);
                }
        }
        public void AddHK() {
            lock(runningArrayLock) foreach(KeyValuePair<IntPtr, List<bool[]>> i in hkRunning) i.Value.Add(new bool[2]);
            hotkeyCount++;
        }
        public void RemoveHK(int index) {
            lock(runningArrayLock) foreach(KeyValuePair<IntPtr, List<bool[]>> j in hkRunning) j.Value.RemoveAt(index);
            hotkeyCount--;
        }

        public void ThreadExitingMacro(MacInfo runInfo) {
            this[runInfo] = false;
            lock(runningArrayLock) if(hkRunning.ContainsKey(runInfo.Handle) && runInfo.MacNumber < hkRunning[runInfo.Handle].Count) hkRunning[runInfo.Handle][runInfo.MacNumber][1] = false;
        }
        public bool IsThreadInMacro(MacInfo runInfo) { lock(runningArrayLock) return !hkRunning.ContainsKey(runInfo.Handle) || runInfo.MacNumber >= hkRunning[runInfo.Handle].Count ? false : hkRunning[runInfo.Handle][runInfo.MacNumber][1]; }

        public void SetToFalse() {
            lock(runningArrayLock)
                foreach(KeyValuePair<IntPtr, List<bool[]>> i in hkRunning)
                    for(byte j = 0; j < i.Value.Count; j++)
                        if(macroOnCount > 0 && i.Value[j][0]) {
                            i.Value[j][0] = false;
                            if(--macroOnCount == 0 && sysTray != null && sysTray.Icon != null) sysTray.Icon = SysNormal;
                        }
        }
        public bool this[MacInfo runInfo] {
            get { lock(runningArrayLock) return !hkRunning.ContainsKey(runInfo.Handle) || runInfo.MacNumber >= hkRunning[runInfo.Handle].Count ? false : hkRunning[runInfo.Handle][runInfo.MacNumber][0]; }
            set {
                lock(runningArrayLock)
                    if(hkRunning.ContainsKey(runInfo.Handle) && runInfo.MacNumber < hkRunning[runInfo.Handle].Count && hkRunning[runInfo.Handle][runInfo.MacNumber][0] != value) {
                        if(value) { hkRunning[runInfo.Handle][runInfo.MacNumber][0] = hkRunning[runInfo.Handle][runInfo.MacNumber][1] = value; macroOnCount++; } else { hkRunning[runInfo.Handle][runInfo.MacNumber][0] = value; macroOnCount--; }

                        if(macroOnCount == 0 && sysTray != null) sysTray.Icon = SysNormal;
                        else if(macroOnCount == 1 && value && sysTray != null) sysTray.Icon = SysRun;
                    }
            }
        }
    }

    public class MacInfo {
        public IntPtr Handle { get; private set; }
        public byte MacNumber { get; private set; }

        public MacInfo(IntPtr handle, byte macNumber) {
            Handle = handle;
            MacNumber = macNumber;
        }

        public bool Equals(MacInfo target) { return MacNumber == target.MacNumber && Handle == target.Handle; }
        public override int GetHashCode() { return Handle.GetHashCode() ^ MacNumber.GetHashCode(); }
        public override bool Equals(object obj) { MacInfo pairObj = obj as MacInfo; return pairObj != null ? Equals(pairObj) : false; }
        public static bool operator ==(MacInfo targetLeft, MacInfo targetRight) { return targetLeft.Equals(targetRight); }
        public static bool operator !=(MacInfo targetLeft, MacInfo targetRight) { return !targetLeft.Equals(targetRight); }
    }

    public class Pointer {
        private List<ushort> offs;
        private Dictionary<string, ushort> extras;

        public uint BasePtr { get; set; }
        public List<ushort> Offs { get { if(offs == null) offs = new List<ushort>(); return offs; } set { offs = value; } }
        public Dictionary<string, ushort> Extras { get { if(extras == null) extras = new Dictionary<string, ushort>(); return extras; } set { extras = value; } }
    }

    public delegate void RunMacro(MacInfo runInfo);

    [Flags] //WinUser.h
    public enum WindowModifiers : uint {
        None = 0x0,
        MOD_ALT = 0x0001,
        MOD_CONTROL = 0x0002,
        MOD_SHIFT = 0x0004,
        MOD_WIN = 0x0008
    }

    [Flags] //DAMac Functions (Others)
    public enum DAMacOps : ushort {
        None = 0x0,
        M_TGTMOUSE = 0x1,
        M_TGTSELF = 0x2,
        M_MINIMODE = 0x4,

        DATK_NOCHKSPELLS = 0x8,
        DATK_NORESTOREPANE = 0x10,
        DATK_WAITFORPANECHANG = 0x20,
        DATK_CLKMOUSEFIRST = 0x40,
        TIME_WAITFORRET = 0x80,

        CM_DBLCLK = 0x100,
        CM_RESTOREPOS = 0x200,
        CM_RCLICK = 0x400,
        CM_MCLICK = 0x800,

        MM_SCREEN = 0x1000
    }
    [Flags] //DAMac Functions (Send Keys)
    public enum DAMacKeyOps : byte {
        None = 0x0,
        SK_REAL = 0x1,
        ST_ENTERS = 0x2
    }

    public enum DASlotSearchType : byte {
        Contains,
        StartsWith,
        EndsWith,
        Exact
    }
}

namespace DATypesLib {
    public enum DAElement : byte {
        None,
        Metal,
        Nature,
        Earth,
        Water,
        Fire,
        Wind,
        Dark,
        Light
    }

    public enum DAPane : byte {
        a = 0, s = 3, S = 4,
        d = 1, D = 2, f = 5,
        F = 6, g = 7, G = 8,
        h = 9
    }

    public enum DAEquipment : byte {
        Weapon = 0,
        Armor = 1,
        Shield = 2,
        Hat = 3,
        Earrings = 4,
        Necklace = 5,
        LRing = 6,
        RRing = 7,
        LGaunt = 8,
        RGaunt = 9,
        Belt = 10,
        Greaves = 11,
        Boots = 12,
        Accessory = 13,
        ShowCloths = 14,
        ShowHat = 15,
        Accessory2 = 16
    }

    public enum DAStat : byte {
        //G Window
        STR = 0,
        INT = 1,
        WIS = 2,
        CON = 3,
        DEX = 4,

        CurHP = 5,
        MaxHP = 6,
        CurMP = 7,
        MaxMP = 8,

        EXP = 9,
        AEXP = 10,
        Gold = 11,

        GP = 12,
        Level = 13,
        NextLEV = 14,
        Ability = 15,
        NextAb = 16,

        //Shift-G Window
        ATK_ELE = 0xF0,
        DEF_ELE = 0xF1,
        MR = 0xF2,
        AC = 0xF3,
        DAM = 0xF4,
        HIT = 0xF5,
    }

    public enum DASpell : byte {
        None,

        Disease = 0x01,
        Mezmerize_WormSkull_Shout = 0x02,
        Blind = 0x03,
        BeagCradh = 0x05,
        EisdCreutair = 0x07,
        Hide = 0x0A,
        NaomhAite = 0x0B,
        CreagNeart = 0x0D,
        Beannaich = 0x10,
        PerfectDefense_WeatherProtection = 0x13,
        Slan_Silence = 0x1A,
        Poison = 0x23,
        Burn_Pause = 0x28,
        Sleep = 0x32,
        FasDeireas = 0x34,
        Dion = 0x35,
        DeireasFaileas_AsgallFaileas = 0x36,
        Mist = 0x37,
        Cradh = 0x52,
        MorCradh = 0x53,
        ArdCradh = 0x54,
        Skull = 0x59,
        Pramh = 0x5A,
        Armachd_AegisSphere = 0x5E,
        BeagSuain = 0x61,
        ClawFist = 0x63,
        WolfFangFist = 0x65,
        FasNadur = 0x77,
        CatsHearing = 0x7C,
        InnerFire = 0x7E,
        DarkSeal = 0x85,
        Decay = 0x8D,
        Purify = 0x8F,
        Regeneration = 0x92,
        CounterAttack = 0x96,
        FeralForm = 0xB7
    }
}