using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;
using System.Xml.Linq;
using DAMacTypesLib;
using DATypesLib;

namespace DAMac {
    public class DAMacCore : IDisposable {
        private IntPtr mainHandle;
        public RunningArray HKRunning { get; private set; }
        private List<Keys> hkKeys = new List<Keys>();
        private List<WindowModifiers> hkMods = new List<WindowModifiers>();
        private List<RunMacro> hkMacros = new List<RunMacro>();
        private List<ushort> hkAtomNames = new List<ushort>();
        private List<bool> hkCheckForDA = new List<bool>();
        private readonly object daMacLock = new object();

        public DAMacCore(IntPtr mainHandle) : this(mainHandle, null, null, null) { }
        public DAMacCore(IntPtr mainHandle, NotifyIcon sysTray, Icon sysNormal, Icon sysRun) {
            this.mainHandle = mainHandle;
            HKRunning = new RunningArray(sysTray, sysNormal, sysRun);
            SetDAPointerInfo("DAData.xml");
        }

        private Dictionary<IntPtr, Process> myProcesses = new Dictionary<IntPtr, Process>();
        private bool StartProcessCheck(IntPtr currProccess) {
            lock(daMacLock)
                if(!myProcesses.ContainsKey(currProccess)) {
                    RefreshAddProcessLists();
                    if(!myProcesses.ContainsKey(currProccess)) return false;
                } else if(myProcesses[currProccess].HasExited) {
                    RemoveProccessLists(currProccess);
                    return false;
                }
            return true;
        }
        private Dictionary<IntPtr, MyMemoryReadSearchLib.MyMemoryReadSearch> pReader = new Dictionary<IntPtr, MyMemoryReadSearchLib.MyMemoryReadSearch>();
        private void RefreshAddProcessLists() {
            foreach(IntPtr i in myProcesses.Keys.ToArray<IntPtr>()) if(myProcesses[i].HasExited) RemoveProccessLists(i);

            Process[] myProcessArray = Process.GetProcessesByName("Darkages");
            for(byte i = 0; i < myProcessArray.Length; i++) {
                if(!daGuiSkillPtr.ContainsKey(myProcessArray[i].MainWindowHandle)) {
                    daGuiSkillPtr.Add(myProcessArray[i].MainWindowHandle, null);
                    daGuiSkill.Add(myProcessArray[i].MainWindowHandle, null);
                }
                if(!pReader.ContainsKey(myProcessArray[i].MainWindowHandle)) {
                    pReader.Add(myProcessArray[i].MainWindowHandle, new MyMemoryReadSearchLib.MyMemoryReadSearch());
                    pReader[myProcessArray[i].MainWindowHandle].OpenProcess((uint)myProcessArray[i].Id);
                }
                if(!myProcesses.ContainsKey(myProcessArray[i].MainWindowHandle)) {
                    myProcesses.Add(myProcessArray[i].MainWindowHandle, myProcessArray[i]);
                    HKRunning.AddProcess(myProcessArray[i].MainWindowHandle);
                } else myProcessArray[i].Dispose();
            }
        }
        private void RemoveProccessLists(IntPtr proccess) {
            myProcesses[proccess].Dispose();
            myProcesses.Remove(proccess);
            HKRunning.RemoveProcess(proccess);
            pReader[proccess].Dispose();
            pReader.Remove(proccess);

            daGuiSkillPtr.Remove(proccess);
            daGuiSkill.Remove(proccess);
        }

        private byte? ReadDAByte(uint lpBaseAddress, MacInfo runInfo) {
            byte[] lpBuffer = null;
            lock(daMacLock) if(pReader.ContainsKey(runInfo.Handle)) lpBuffer = pReader[runInfo.Handle].ReadProcessMemory((IntPtr)lpBaseAddress, (UIntPtr)1);

            if(lpBuffer == null) { HKRunning[runInfo] = false; return null; }
            return lpBuffer[0];
        }
        private ushort? ReadDAUShort(uint lpBaseAddress, MacInfo runInfo) {
            byte[] lpBuffer = null;
            lock(daMacLock) if(pReader.ContainsKey(runInfo.Handle)) lpBuffer = pReader[runInfo.Handle].ReadProcessMemory((IntPtr)lpBaseAddress, (UIntPtr)2);

            if(lpBuffer == null) { HKRunning[runInfo] = false; return null; }
            return (ushort)(lpBuffer[0] + 256 * lpBuffer[1]);
        }
        private uint? ReadDAUInt(uint lpBaseAddress, MacInfo runInfo) { return ReadDAUInt(lpBaseAddress, runInfo, true); }
        private uint? ReadDAUInt(uint lpBaseAddress, MacInfo runInfo, bool setToFalse) {
            byte[] lpBuffer = null;
            lock(daMacLock) if(pReader.ContainsKey(runInfo.Handle)) lpBuffer = pReader[runInfo.Handle].ReadProcessMemory((IntPtr)lpBaseAddress, (UIntPtr)4);

            if(lpBuffer == null) { if(setToFalse) HKRunning[runInfo] = false; return null; }
            return (uint)(lpBuffer[0] + 256 * lpBuffer[1] + 256 * 256 * lpBuffer[2] + 256 * 256 * 256 * lpBuffer[3]);
        }
        private double? ReadDAGeneric(uint lpBaseAddress, byte nSize, MacInfo runInfo) {
            byte[] lpBuffer = null;
            lock(daMacLock) if(pReader.ContainsKey(runInfo.Handle)) lpBuffer = pReader[runInfo.Handle].ReadProcessMemory((IntPtr)lpBaseAddress, (UIntPtr)nSize);

            if(lpBuffer == null) { HKRunning[runInfo] = false; return null; }
            double ret = lpBuffer[0];
            for(byte i = 1; i < nSize; i++) ret += Math.Pow(256, (double)i) * lpBuffer[i];
            return ret;
        }
        private uint? ReadPointer(Pointer pointer, MacInfo runInfo) { return ReadPointer(pointer, runInfo, true); }
        private uint? ReadPointer(Pointer pointer, MacInfo runInfo, bool setToFalse) {
            uint? address = pointer.BasePtr;
            foreach(ushort A in pointer.Offs)
                if(!(address = ReadDAUInt(address.Value, runInfo, setToFalse)).HasValue) return null;
                else address += A;
            return address;
        }

        private readonly object useDAAttackLock = new object();
        private Keys[] mapDAHotKeys = new Keys[] { Keys.None, Keys.D1, Keys.D2, Keys.D3, Keys.D4, Keys.D5, Keys.D6, Keys.D7, Keys.D8, Keys.D9, Keys.D0, Keys.OemMinus, Keys.Oemplus };
        public void UseDAAttack(byte targetSlot, DAPane targetPane, bool mouseClick, MacInfo runInfo) { UseDAAttack(new byte[] { targetSlot }, new DAPane[] { targetPane }, new bool[] { mouseClick }, runInfo, null, null, 0); }
        public void UseDAAttack(byte[] targetSlot, DAPane[] targetPane, bool[] mouseClick, MacInfo runInfo) { UseDAAttack(targetSlot, targetPane, mouseClick, runInfo, null, null, 0); }
        public void UseDAAttack(byte targetSlot, DAPane targetPane, bool mouseClick, MacInfo runInfo, DAMacOps option) { UseDAAttack(new byte[] { targetSlot }, new DAPane[] { targetPane }, new bool[] { mouseClick }, runInfo, null, null, option); }
        public void UseDAAttack(byte[] targetSlot, DAPane[] targetPane, bool[] mouseClick, MacInfo runInfo, DAMacOps option) { UseDAAttack(targetSlot, targetPane, mouseClick, runInfo, null, null, option); }
        public void UseDAAttack(byte targetSlot, DAPane targetPane, bool mouseClick, MacInfo runInfo, ushort? locX, ushort? locY) { UseDAAttack(new byte[] { targetSlot }, new DAPane[] { targetPane }, new bool[] { mouseClick }, runInfo, locX, locY, 0); }
        public void UseDAAttack(byte[] targetSlot, DAPane[] targetPane, bool[] mouseClick, MacInfo runInfo, ushort? locX, ushort? locY) { UseDAAttack(targetSlot, targetPane, mouseClick, runInfo, locX, locY, 0); }
        public void UseDAAttack(byte targetSlot, DAPane targetPane, bool mouseClick, MacInfo runInfo, ushort? locX, ushort? locY, DAMacOps option) { UseDAAttack(new byte[] { targetSlot }, new DAPane[] { targetPane }, new bool[] { mouseClick }, runInfo, locX, locY, option); }
        public void UseDAAttack(byte[] targetSlot, DAPane[] targetPane, bool[] mouseClick, MacInfo runInfo, ushort? locX, ushort? locY, DAMacOps option) {
            Monitor.Enter(useDAAttackLock);
            try {
                //Suain(SleepGraphic) is Buged... The graphic is on you sometimes when your not actually suained (can't remove graphic)
                if((option & DAMacOps.DATK_NOCHKSPELLS) == 0) { if(IsDASpellIconOn(new DASpell[] { DASpell.Sleep, DASpell.Pramh, DASpell.WolfFangFist, DASpell.Skull, DASpell.Mezmerize_WormSkull_Shout }, runInfo)) return; }
                if((option & DAMacOps.DATK_CLKMOUSEFIRST) != 0) ClickDAMouse(runInfo, 10, 470, DAMacOps.CM_RESTOREPOS);

                DAPane? currPane = GetDAPane(runInfo);
                if(!currPane.HasValue) return;

                bool daMinMode = IsDAPaneMinimized(runInfo);
                bool daExtended = IsDAPaneExtended(runInfo);
                byte maxLoc = targetSlot.Max<byte>();

                bool restoreExtendState = false;
                if(maxLoc > 12 && daMinMode && !daExtended) {
                    restoreExtendState = true;
                    ToggleExtend(currPane.Value, runInfo);
                    daExtended = !daExtended;
                } else if(!daMinMode && daExtended && !(targetPane[targetPane.Length - 1] == DAPane.a && targetSlot[targetSlot.Length - 1] > 36))
                    for(byte i = 0; i < targetPane.Length; i++) if(targetPane[i] != DAPane.a) { restoreExtendState = true; break; }

                bool restorePane = false;
                if((option & DAMacOps.DATK_NORESTOREPANE) == 0 && targetPane[targetPane.Length - 1] != currPane) restorePane = true;

                bool spellRestiction = true;
                for(byte i = 0; i < targetPane.Length; i++) if(targetPane[i] != DAPane.d && targetPane[i] != DAPane.D) { spellRestiction = false; break; }

                bool restoreMouse = false;
                Point currMousePos = GetDAMousePos(runInfo);
                if(maxLoc > 12 || (option & DAMacOps.M_TGTSELF) != 0 || (option & DAMacOps.M_TGTMOUSE) != 0) restoreMouse = true;

                for(byte i = 0; i < targetSlot.Length; i++) {
                    if(i > 0) {
                        if(targetPane[i - 1] != targetPane[i]) {
                            SwitchPane(targetPane[i], runInfo);
                            if(!daMinMode && daExtended) daExtended = !daExtended;
                        }
                    } else if(currPane != targetPane[i]) {
                        SwitchPane(targetPane[i], runInfo);
                        if(!daMinMode && daExtended) daExtended = !daExtended;
                    }

                    if(targetSlot[i] < 13) SendKeyPM(mapDAHotKeys[targetSlot[i]], runInfo.Handle);
                    else ClickDAAttack(targetSlot[i], targetPane[i], ref daMinMode, ref daExtended, runInfo);

                    if(mouseClick[i])
                        if((option & DAMacOps.M_TGTMOUSE) != 0) ClickDAMouse(runInfo, DAMacOps.M_TGTMOUSE);
                        else if((option & DAMacOps.M_TGTSELF) != 0) {
                            if(daMinMode) ClickDAMouse(runInfo, DAMacOps.M_TGTSELF | DAMacOps.M_MINIMODE);
                            else ClickDAMouse(runInfo, DAMacOps.M_TGTSELF);
                        } else ClickDAMouse(runInfo, locX, locY);

                    if(spellRestiction && i % 3 == 2 && i != targetSlot.Length - 1) {
                        if(restoreExtendState) ToggleExtend(targetPane[i], runInfo);
                        Thread.Sleep(1025);
                        if(restoreExtendState) ToggleExtend(targetPane[i], runInfo);
                    }
                }

                if(restoreExtendState && restorePane) {
                    ToggleExtend(currPane.Value, runInfo);
                    if(currPane != DAPane.a && currPane != DAPane.h) SwitchPane(currPane.Value, runInfo);
                } else if(restoreExtendState && !restorePane) ToggleExtend(targetPane[targetPane.Length - 1], runInfo);
                else if(restorePane) SwitchPane(currPane.Value, runInfo);

                if(restoreMouse) MoveDAMouse(runInfo, (ushort)currMousePos.X, (ushort)currMousePos.Y);
            } finally { Thread.Sleep(5); Monitor.Exit(useDAAttackLock); }
        }
        private void ClickDAAttack(byte targetSlot, DAPane targetPane, ref bool daMinMode, ref bool daExtended, MacInfo runInfo) {
            ushort yCoordSkill, xCoordSkill;
            byte row = (byte)((targetSlot - 1) / 12);

            if(row > 2 && !daExtended) {
                ToggleExtend(targetPane, runInfo);
                daExtended = !daExtended;
            }

            // Y = (Row)*SingleRowOffset + SkillWin2FirstRow + Top2SkillWin
            if(daMinMode && targetPane == DAPane.a) yCoordSkill = (ushort)(row * 32 + 16 + 312);
            else if(daMinMode) yCoordSkill = (ushort)(row * 32 + 16 + 378);
            else if(!daMinMode && daExtended) yCoordSkill = (ushort)(row * 32 + 16 + 269);
            else yCoordSkill = (ushort)(row * 32 + 16 + 338);
            // X = (Subtract Excess Rows, and Sub 1)*SingleSkillOffset + SkillWin2FirstSkill + Edge2SkillWin
            xCoordSkill = (ushort)((targetSlot - 1 - 12 * row) * 36 + 18 + 95);

            ClickDAMouse(runInfo, xCoordSkill, yCoordSkill, DAMacOps.CM_DBLCLK);
        }
        private void ToggleExtend(DAPane pane, MacInfo runInfo) {
            switch(pane) {
                case DAPane.a: SendKeyPM(Keys.A, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.s: SendKeyPM(Keys.S, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.S: SendKeyPM(Keys.S, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.d: SendKeyPM(Keys.D, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.D: SendKeyPM(Keys.D, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.f: SendKeyPM(Keys.F, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.F: SendKeyPM(Keys.F, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.g: SendKeyPM(Keys.G, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.G: SendKeyPM(Keys.G, Keys.ControlKey, runInfo.Handle); break;
                case DAPane.h: SendKeyPM(Keys.H, Keys.ControlKey, runInfo.Handle); break;
            }
        }
        private void SwitchPane(DAPane pane, MacInfo runInfo) {
            switch(pane) {
                case DAPane.a: SendKeyPM(Keys.A, runInfo.Handle); break;
                case DAPane.s: SendKeyPM(Keys.S, runInfo.Handle); break;
                case DAPane.S: SendKeyPM(Keys.S, Keys.ShiftKey, runInfo.Handle); break;
                case DAPane.d: SendKeyPM(Keys.D, runInfo.Handle); break;
                case DAPane.D: SendKeyPM(Keys.D, Keys.ShiftKey, runInfo.Handle); break;
                case DAPane.f: SendKeyPM(Keys.F, runInfo.Handle); break;
                case DAPane.F: SendKeyPM(Keys.F, Keys.ShiftKey, runInfo.Handle); break;
                case DAPane.g: SendKeyPM(Keys.G, runInfo.Handle); break;
                case DAPane.G: SendKeyPM(Keys.G, Keys.ShiftKey, runInfo.Handle); break;
                case DAPane.h: SendKeyPM(Keys.H, runInfo.Handle); break;
            }
        }

        public void SendTextMsgPM(string send, IntPtr daWin) { SendTextMsgPM(send, daWin, 0); }
        public void SendTextMsgPM(string send, IntPtr daWin, DAMacKeyOps option) {
            if((option & DAMacKeyOps.ST_ENTERS) != 0) {
                SendKeyPM(Keys.Enter, daWin, DAMacKeyOps.SK_REAL);
                Thread.Sleep(85);
            }
            foreach(char sendChar in send.ToCharArray()) SafeNativeMethods.PostMessage(daWin, SafeNativeMethods.WindowMessage.WM_CHAR, (UIntPtr)sendChar, (UIntPtr)SafeNativeMethods.MAKELONG(0x1, (ushort)SafeNativeMethods.MapVirtualKey(SafeNativeMethods.LOBYTE((ushort)SafeNativeMethods.VkKeyScan(sendChar)), 0)));
            if((option & DAMacKeyOps.ST_ENTERS) != 0) SendKeyPM(Keys.Enter, daWin);
        }

        public void SendKeysPM(string send, IntPtr daWin) { SendKeysPM(send, daWin, 0); }
        public void SendKeysPM(string send, IntPtr daWin, DAMacKeyOps option) {
            ushort vk;
            char[] sendChars = send.ToCharArray();
            for(ushort i = 0; i < sendChars.Length; i++) {
                vk = (ushort)SafeNativeMethods.VkKeyScan(sendChars[i]);
                switch(SafeNativeMethods.HIBYTE(vk)) {
                    case 0: SendKeyPM(SafeNativeMethods.LOBYTE(vk), null, daWin, option); break;
                    case 1: SendKeyPM(SafeNativeMethods.LOBYTE(vk), new uint[] { (uint)Keys.ShiftKey }, daWin, option); break;
                    case 2: SendKeyPM(SafeNativeMethods.LOBYTE(vk), new uint[] { (uint)Keys.ControlKey }, daWin, option); break;
                    case 3: SendKeyPM(SafeNativeMethods.LOBYTE(vk), new uint[] { (uint)Keys.Menu }, daWin, option); break;
                }
            }
        }

        private uint[] extKeys = new uint[] { (uint)Keys.RMenu, (uint)Keys.RControlKey, (uint)Keys.Insert, (uint)Keys.Delete, (uint)Keys.Home, (uint)Keys.End, (uint)Keys.PageUp, (uint)Keys.PageDown, (uint)Keys.Up, (uint)Keys.Down, (uint)Keys.Left, (uint)Keys.Right, (uint)Keys.Divide, (uint)Keys.NoName };
        public void SendKeyPM(Keys key, IntPtr daWin) { SendKeyPM((uint)key, null, daWin, 0); }
        public void SendKeyPM(Keys key, IntPtr daWin, DAMacKeyOps option) { SendKeyPM((uint)key, null, daWin, option); }
        public void SendKeyPM(Keys key, Keys modifier, IntPtr daWin) { SendKeyPM((uint)key, new uint[] { (uint)modifier }, daWin, 0); }
        public void SendKeyPM(Keys key, Keys[] modifier, IntPtr daWin) { SendKeyPM(key, modifier, daWin, 0); }
        public void SendKeyPM(Keys key, Keys modifier, IntPtr daWin, DAMacKeyOps option) { SendKeyPM((uint)key, new uint[] { (uint)modifier }, daWin, option); }
        public void SendKeyPM(Keys key, Keys[] modifier, IntPtr daWin, DAMacKeyOps option) {
            if(modifier == null) SendKeyPM((uint)key, null, daWin, option);
            else {
                uint[] modifersUI = new uint[modifier.Length];
                for(byte i = 0; i < modifier.Length; i++) modifersUI[i] = (uint)modifier[i];
                SendKeyPM((uint)key, modifersUI, daWin, option);
            }
        }
        public void SendKeyPM(uint key, uint[] modifier, IntPtr daWin, DAMacKeyOps option) {
            if((option & DAMacKeyOps.SK_REAL) != 0 && modifier != null) {
                byte[] kbState = new byte[255];

                uint extProcID;
                if(SafeNativeMethods.GetWindowThreadProcessId(daWin, out extProcID) == 0) return;

                if(!SafeNativeMethods.AttachThreadInput((uint)AppDomain.GetCurrentThreadId(), extProcID, true) || !SafeNativeMethods.GetKeyboardState(kbState)) return;
                byte[] kbStateMOD = kbState;
                for(byte i = 0; i < modifier.Length; i++) kbStateMOD[modifier[i]] = 0x80;

                if(!SafeNativeMethods.SetKeyboardState(kbStateMOD)) return;
                SendKeyPMHelp(ref key, ref modifier, ref daWin, ref option);
                if(!SafeNativeMethods.SetKeyboardState(kbState) || !SafeNativeMethods.AttachThreadInput((uint)AppDomain.GetCurrentThreadId(), extProcID, false)) return;
            } else SendKeyPMHelp(ref key, ref modifier, ref daWin, ref option);
        }
        private void SendKeyPMHelp(ref uint key, ref uint[] modifier, ref IntPtr daWin, ref DAMacKeyOps option) {
            ushort extKey = 0x0000;
            if(extKeys.Contains(key)) {
                extKey = 0x0100;
                if(key == (uint)Keys.RMenu) key = (uint)Keys.Menu;
                else if(key == (uint)Keys.RControlKey) key = (uint)Keys.ControlKey;
                else if(key == (uint)Keys.NoName) key = (uint)Keys.Enter;
            }

            if(modifier != null)
                for(byte i = 0; i < modifier.Length; i++) {
                    if(modifier[i] == (uint)Keys.Menu) SafeNativeMethods.PostMessage(daWin, SafeNativeMethods.WindowMessage.WM_SYSKEYDOWN, (UIntPtr)modifier[i], (UIntPtr)SafeNativeMethods.MAKELONG(0x1, (ushort)(SafeNativeMethods.MapVirtualKey(modifier[i], 0) | extKey)));
                    else SafeNativeMethods.PostMessage(daWin, SafeNativeMethods.WindowMessage.WM_KEYDOWN, (UIntPtr)modifier[i], (UIntPtr)SafeNativeMethods.MAKELONG(0x1, (ushort)(SafeNativeMethods.MapVirtualKey(modifier[i], 0) | extKey)));
                }

            SafeNativeMethods.PostMessage(daWin, SafeNativeMethods.WindowMessage.WM_KEYDOWN, (UIntPtr)key, (UIntPtr)SafeNativeMethods.MAKELONG(0x1, (ushort)(SafeNativeMethods.MapVirtualKey(key, 0) | extKey)));
            if((option & DAMacKeyOps.SK_REAL) != 0) Thread.Sleep(10);
            SafeNativeMethods.PostMessage(daWin, SafeNativeMethods.WindowMessage.WM_KEYUP, (UIntPtr)key, (UIntPtr)SafeNativeMethods.MAKELONG(0x1, (ushort)(SafeNativeMethods.MapVirtualKey(key, 0) | 0xC000 | extKey)));  //Restores Key State (Not Really Needed but simulates "real" more
            if(modifier != null)
                for(byte i = 0; i < modifier.Length; i++) {
                    if(modifier[i] == (uint)Keys.Menu) SafeNativeMethods.PostMessage(daWin, SafeNativeMethods.WindowMessage.WM_SYSKEYUP, (UIntPtr)modifier[i], (UIntPtr)SafeNativeMethods.MAKELONG(0x1, (ushort)(SafeNativeMethods.MapVirtualKey(key, 0) | 0xC000 | extKey)));
                    else SafeNativeMethods.PostMessage(daWin, SafeNativeMethods.WindowMessage.WM_KEYUP, (UIntPtr)modifier[i], (UIntPtr)SafeNativeMethods.MAKELONG(0x1, (ushort)(SafeNativeMethods.MapVirtualKey(modifier[i], 0) | 0xC000 | extKey)));
                }
        }

        public void ClickDAMouse(MacInfo runInfo, ushort? locX, ushort? locY) { ClickDAMouse(runInfo, locX, locY, 0); }
        public void ClickDAMouse(MacInfo runInfo, DAMacOps option) { ClickDAMouse(runInfo, null, null, option); }
        public void ClickDAMouse(MacInfo runInfo, ushort? locX, ushort? locY, DAMacOps option) {
            if((option & DAMacOps.CM_RESTOREPOS) != 0) {
                Point mouseCoordRestore = GetDAMousePos(runInfo);
                ClickDAMouseHelp(runInfo, locX, locY, option);
                MoveDAMouse(runInfo, (ushort)mouseCoordRestore.X, (ushort)mouseCoordRestore.Y);
            } else ClickDAMouseHelp(runInfo, locX, locY, option);
        }
        private void ClickDAMouseHelp(MacInfo runInfo, ushort? locX, ushort? locY, DAMacOps option) {
            if((option & DAMacOps.M_TGTMOUSE) != 0) {
                Point mouseCoord = GetDAMousePos(runInfo);
                locX = (ushort)mouseCoord.X; locY = (ushort)mouseCoord.Y;
            } else {
                if((option & DAMacOps.M_TGTSELF) != 0) {
                    locX = 311; locY = 145;
                    if((option & DAMacOps.M_MINIMODE) != 0) locY += 30;
                } else if(!locX.HasValue || !locY.HasValue) return;
            }

            UIntPtr lParam = (UIntPtr)SafeNativeMethods.MAKELONG(locX.Value, locY.Value), wParam = (UIntPtr)1;
            SafeNativeMethods.WindowMessage buttonDown = SafeNativeMethods.WindowMessage.WM_LBUTTONDOWN, buttonUp = SafeNativeMethods.WindowMessage.WM_LBUTTONUP;
            if((option & DAMacOps.CM_RCLICK) != 0) {
                wParam = (UIntPtr)0x02;
                buttonDown = SafeNativeMethods.WindowMessage.WM_RBUTTONDOWN;
                buttonUp = SafeNativeMethods.WindowMessage.WM_RBUTTONUP;
            } else if((option & DAMacOps.CM_MCLICK) != 0) {
                wParam = (UIntPtr)0x10;
                buttonDown = SafeNativeMethods.WindowMessage.WM_MBUTTONDOWN;
                buttonUp = SafeNativeMethods.WindowMessage.WM_MBUTTONUP;
            }

            MoveDAMouse(runInfo, locX.Value, locY.Value);  //Note: Should only have to do this if not M_TGTMOUSE, but if ClickDAMouse is turning on spell, then DAMouse wont be in right spot and spell wont cast.
            SafeNativeMethods.PostMessage(runInfo.Handle, buttonDown, wParam, lParam);
            SafeNativeMethods.PostMessage(runInfo.Handle, buttonUp, UIntPtr.Zero, lParam);
            if((option & DAMacOps.CM_DBLCLK) != 0) {
                SafeNativeMethods.PostMessage(runInfo.Handle, buttonDown, wParam, lParam);
                SafeNativeMethods.PostMessage(runInfo.Handle, buttonUp, UIntPtr.Zero, lParam);
            }
        }

        public void MoveDAMouse(MacInfo runInfo, DAMacOps option) { MoveDAMouse(runInfo, null, null, option); }
        public void MoveDAMouse(MacInfo runInfo, ushort? locX, ushort? locY) { MoveDAMouse(runInfo, locX, locY, 0); }
        private void MoveDAMouse(MacInfo runInfo, ushort? locX, ushort? locY, DAMacOps option) {
            if((option & DAMacOps.M_TGTSELF) != 0) {
                locX = 311; locY = 145;
                if((option & DAMacOps.M_MINIMODE) != 0) locY += 30;
            } else if(!locX.HasValue || !locY.HasValue) return;
            SafeNativeMethods.PostMessage(runInfo.Handle, SafeNativeMethods.WindowMessage.WM_MOUSEMOVE, UIntPtr.Zero, (UIntPtr)SafeNativeMethods.MAKELONG(locX.Value, locY.Value));
        }

        public void MoveMouse(MacInfo runInfo, DAMacOps option) { MoveMouse(runInfo, null, null, option); }
        public void MoveMouse(MacInfo runInfo, ushort? locX, ushort? locY) { MoveMouse(runInfo, locX, locY, 0); }
        private void MoveMouse(MacInfo runInfo, ushort? locX, ushort? locY, DAMacOps option) {
            Point mousePoint;
            if((option & DAMacOps.M_TGTSELF) != 0) {
                mousePoint = new Point(311, 145);
                if((option & DAMacOps.M_MINIMODE) != 0) mousePoint.Y += 30;
            } else if(locX.HasValue && locY.HasValue) mousePoint = new Point(locX.Value, locY.Value);
            else return;

            if((option & DAMacOps.MM_SCREEN) == 0) SafeNativeMethods.ScreenToClient(runInfo.Handle, ref mousePoint);
            Cursor.Position = mousePoint;
        }

        //NOTE: Wont work if DA Is Minimized.  Returns ARGB Format
        public uint GetPixelColor(int locX, int locY) { return GetPixelColor(new Point(locX, locY), IntPtr.Zero); }
        public uint GetPixelColor(Point daCoordinates) { return GetPixelColor(daCoordinates, IntPtr.Zero); }
        public uint GetPixelColor(int locX, int locY, IntPtr daWin) { return GetPixelColor(new Point(locX, locY), daWin); }
        public uint GetPixelColor(Point daCoordinates, IntPtr daWin) {
            if(daWin != IntPtr.Zero) SafeNativeMethods.ClientToScreen(daWin, ref daCoordinates);
            if(daCoordinates.X < 0 || daCoordinates.Y < 0) return 0;

            using(Bitmap daImage = new Bitmap(1, 1)) {
                using(Graphics DAWindowGraphic = Graphics.FromImage(daImage)) DAWindowGraphic.CopyFromScreen(daCoordinates.X, daCoordinates.Y, 0, 0, new Size(1, 1));
                return (uint)daImage.GetPixel(0, 0).ToArgb();
            }
        }
        public Point GetDAMousePos(MacInfo runInfo) {
            Point mouseCoord = Cursor.Position;
            SafeNativeMethods.ScreenToClient(runInfo.Handle, ref mouseCoord);
            return mouseCoord;
        }
        private string GetDAString(uint address, MacInfo runInfo) {
            byte? resultT;
            StringBuilder result = new StringBuilder();

            for(uint i = address; (resultT = ReadDAByte(i, runInfo)) != 0; i++) {
                if(!resultT.HasValue) return "";
                result.Append((char)resultT);
            }
            return result.ToString();
        }

        public string GetDAName(MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["Name"], runInfo);
            return !address.HasValue ? "" : GetDAString(address.Value, runInfo);
        }
        public string GetDAEquipment(DAEquipment daEquipment, MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["Equip"], runInfo);
            return !address.HasValue ? "" : GetDAString(address.Value + (uint)daEquipment * daPtrs["GWin"].Extras["size"], runInfo);
        }
        public string GetDAStatAscii(DAStat daStat, MacInfo runInfo) {
            bool sGwin = daStat == DAStat.ATK_ELE || daStat == DAStat.DEF_ELE || daStat == DAStat.MR || daStat == DAStat.AC || daStat == DAStat.DAM || daStat == DAStat.HIT;
            uint? address = sGwin ? ReadPointer(daPtrs["SGWin"], runInfo) : ReadPointer(daPtrs["GWin"], runInfo);
            return !address.HasValue ? "" : GetDAString(address.Value + ((uint)daStat - (sGwin ? 0xF0u : 0x00u)) * (sGwin ? daPtrs["SGWin"].Extras["size"] : daPtrs["GWin"].Extras["size"]), runInfo);
        }
        public uint GetDAStatNumber(DAStat daStat, MacInfo runInfo) {
            uint result;
            return UInt32.TryParse(GetDAStatAscii(daStat, runInfo), out result) ? result : 0xFFFFFFFF;
        }
        public byte[] GetDACharCoordinates(MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["Coords"], runInfo);
            if(!address.HasValue) return null;

            byte? currX = ReadDAByte(address.Value + daPtrs["Coords"].Extras["size"], runInfo), currY = ReadDAByte(address.Value, runInfo);
            return !currX.HasValue || !currY.HasValue ? null : new byte[] { currX.Value, currY.Value };
        }
        public ushort? GetDAMapNumber(MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["MapNum"], runInfo);
            return !address.HasValue ? null : ReadDAUShort(address.Value, runInfo);
        }
        public DAPane? GetDAPane(MacInfo runInfo) {
            byte? pane;
            uint? address = ReadPointer(daPtrs["Panel"], runInfo);
            return !address.HasValue ? null : ((pane = ReadDAByte(address.Value, runInfo)) >= 0 && pane < 10 ? (DAPane?)pane : null);
        }
        public bool IsDAPaneExtended(MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["Extend"], runInfo);
            return !address.HasValue ? false : ReadDAByte(address.Value, runInfo) == 1;
        }
        public bool IsDAPaneMinimized(MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["Mini"], runInfo);
            return !address.HasValue ? false : ReadDAByte(address.Value, runInfo) == 1;
        }
        public bool IsDASpellIconOn(DASpell daSpell, MacInfo runInfo) { return IsDASpellIconOn(new DASpell[] { daSpell }, runInfo); }
        public bool IsDASpellIconOn(DASpell[] daSpell, MacInfo runInfo) {
            uint? addressOffset = ReadPointer(daPtrs["SpellOff"], runInfo);
            if(!addressOffset.HasValue || !(addressOffset = ReadDAUShort(addressOffset.Value, runInfo)).HasValue) return false;

            Pointer tmpSpellPtr = new Pointer() { BasePtr = daPtrs["Spell"].BasePtr, Offs = new List<ushort>(daPtrs["Spell"].Offs) };
            tmpSpellPtr.Offs.Insert(0, (ushort)(daPtrs["Spell"].Extras["o1size"] * addressOffset.Value + daPtrs["Spell"].Extras["o1off1"]));

            //Keep trying the o1off# till one works
            uint? address = ReadPointer(tmpSpellPtr, runInfo, false);
            if(!address.HasValue || (address.HasValue && address.Value <= 0x4000)) {
                for(byte i = 2; daPtrs["Spell"].Extras.ContainsKey("o1off" + i.ToString()); i++) {
                    tmpSpellPtr.Offs[0] = (ushort)(daPtrs["Spell"].Extras["o1size"] * addressOffset.Value + daPtrs["Spell"].Extras["o1off" + i.ToString()]);
                    address = ReadPointer(tmpSpellPtr, runInfo, false);
                    if(address.HasValue && address.Value > 0x4000) break;
                }
                if(!address.HasValue || (address.HasValue && address.Value <= 0x4000)) { HKRunning[runInfo] = false; return false; }
            }

            for(byte i = 0; i < 10; i++) foreach(DASpell A in daSpell) if(ReadDAByte(address.Value + (byte)(i * daPtrs["Spell"].Extras["o2size"]), runInfo) == (byte)A) return true;
            return false;
        }

        //Doesn't go to 0 if "a,a" window.  Doesn't go 0 if MessageWindow Up (IE Sence Window / Group Accept / Exchant Window).  Game windows / UI windows seem to go to 0.
        //NOTE: This is usally set to correct values, but some towns (Like Mileth/Oren) make this go to 0 when you enter the map (NB: can press "q,q" to manually set pointer if this is 0, or suspect it might be 0).
        public bool IsDAWindowUp(MacInfo runInfo) { return IsDAWindowUpGetAd(runInfo) == 0; }
        //If the Ad this returns ever changes then that means a window went up (or went off if you started it when a window is already up).  Works for most Windows (not "a,a" or "e", does for "w").  If macro is fast (IE SenceWindowMacro) it will be 100% fine.  If macro is long (IE Macroing Spell) then pressing "q,q" (really quick) sometimes give this a new address value; even if the macro doesn't catch the "q" window while its up, it will probably still detect the change.  Only "q,q" has the habbit of changing the value.
        //Sometimes normal things also change the value, but if you walk through a map good chance it'll go to the way it usally is (non changing value for windows), "q,q" doesn't have this property.
        public uint? IsDAWindowUpGetAd(MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["WinOn"], runInfo, false);
            return address.HasValue && address > 0x4000 ? address.Value : 0;
            //return address.HasValue && address > 0x4000 ? ReadDAUInt(address.Value, runInfo) : 0;
        }
        public string GetDALastWindowText(MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["SenseWin"], runInfo);
            return !address.HasValue ? "" : GetDAString(address.Value, runInfo);
        }
        public string UseGetDALastSenseWindow(byte targetSlot, MacInfo runInfo) {
            uint? windowOrigAd = IsDAWindowUpGetAd(runInfo);
            if(!windowOrigAd.HasValue) return "";

            UseDAAttack(targetSlot, DAPane.s, false, runInfo);

            for(byte i = 0; i < 0xFF; i++) {
                if(IsDAWindowUpGetAd(runInfo) != windowOrigAd) {
                    SendKeyPM(Keys.Escape, runInfo.Handle);
                    try { return GetDALastWindowText(runInfo); } finally { Thread.Sleep(10); SendKeyPM(Keys.Escape, runInfo.Handle); }
                }
                Thread.Sleep(5);
            }
            return "";
        }

        private Dictionary<IntPtr, uint?> daGuiSkillPtr = new Dictionary<IntPtr, uint?>();
        private Dictionary<IntPtr, uint?> daGuiSkill = new Dictionary<IntPtr, uint?>();
        public bool IsDASkillActive(byte slotNumber, DAPane skillPane, MacInfo runInfo) {
            if(skillPane == DAPane.S) slotNumber += 36;
            else if(skillPane == DAPane.h)
                if(slotNumber <= 12) slotNumber += 72;
                else if(slotNumber <= 24) slotNumber += 66;
                else slotNumber += 60;

            lock(daMacLock) {
                if(!SetDASkillPtr(runInfo)) return false;
                Pointer tmpSkillPtr = new Pointer() { BasePtr = daGuiSkill[runInfo.Handle].Value + (byte)(--slotNumber * daPtrs["SkillSrch"].Extras["o1size"]), Offs = daPtrs["SkillSrch"].Offs };
                uint? address = ReadPointer(tmpSkillPtr, runInfo);
                return !address.HasValue ? false : ReadDAByte(address.Value, runInfo) == 1;
            }
        }
        //Have to Manually Lock daMacLock for this (Reasioning: If your settings, you probably want to use afterwards anyway)
        private bool SetDASkillPtr(MacInfo runInfo) {
            if(daGuiSkillPtr.ContainsKey(runInfo.Handle) && daGuiSkillPtr[runInfo.Handle].HasValue && ReadDAUInt(daGuiSkillPtr[runInfo.Handle].Value, runInfo, false) == daPtrs["SkillSrch"].BasePtr && ReadDAUInt(daGuiSkillPtr[runInfo.Handle].Value + daPtrs["SkillSrch"].Extras["o0off"], runInfo, false) == daGuiSkill[runInfo.Handle]) return true;
            if(!(daGuiSkillPtr[runInfo.Handle] = pReader[runInfo.Handle].SearchDAProcessMemory(daPtrs["SkillSrch"].BasePtr)).HasValue) { HKRunning[runInfo] = false; return false; }
            return (daGuiSkill[runInfo.Handle] = ReadDAUInt(daGuiSkillPtr[runInfo.Handle].Value + daPtrs["SkillSrch"].Extras["o0off"], runInfo, true)).HasValue;
        }

        //Spell/Skill Level is in Names, Amount is not in the Invenotry Name.
        public string GetDASlotInformation(DAPane daPaneSearch, byte slotNumber, MacInfo runInfo) {
            uint? address = ReadPointer(daPtrs["Slots"], runInfo);
            if(!address.HasValue) return "";

            uint off = 0;
            if(daPaneSearch == DAPane.a) return GetDAString(address.Value + (uint)--slotNumber * daPtrs["Slots"].Extras["itSize"], runInfo);
            else off = 60u * daPtrs["Slots"].Extras["itSize"];

            if(daPaneSearch == DAPane.d) return GetDAString(address.Value + off + (uint)--slotNumber * daPtrs["Slots"].Extras["spSize"], runInfo);
            else off += 36u * daPtrs["Slots"].Extras["spSize"];
            if(daPaneSearch == DAPane.D) return GetDAString(address.Value + off + (uint)--slotNumber * daPtrs["Slots"].Extras["spSize"], runInfo);
            else off += 36u * daPtrs["Slots"].Extras["spSize"];
            if(daPaneSearch == DAPane.h && ((slotNumber < 13 && slotNumber > 6) || (slotNumber < 25 && slotNumber > 18) || (slotNumber > 30 && slotNumber < 36)))
                return GetDAString(address.Value + off + (uint)--slotNumber * daPtrs["Slots"].Extras["spSize"], runInfo);
            else off += 17u * daPtrs["Slots"].Extras["spSize"] - 1;

            if(daPaneSearch == DAPane.s) return GetDAString(address.Value + off + (uint)--slotNumber * daPtrs["Slots"].Extras["skSize"], runInfo);
            else off += 36u * daPtrs["Slots"].Extras["skSize"];
            if(daPaneSearch == DAPane.S) return GetDAString(address.Value + off + (uint)--slotNumber * daPtrs["Slots"].Extras["skSize"], runInfo);
            else off += 36u * daPtrs["Slots"].Extras["skSize"];
            if(daPaneSearch == DAPane.h && (slotNumber < 7 || (slotNumber < 19 && slotNumber > 12) || (slotNumber < 30 && slotNumber > 24)))
                return GetDAString(address.Value + off + (uint)--slotNumber * daPtrs["Slots"].Extras["skSize"], runInfo);
            else return "";
        }
        //Returns an empty array if nothing found.
        public byte[] GetDASlot(DAPane searchPane, string itemToSearch, bool returnAll, MacInfo runInfo, DASlotSearchType searchType) { return GetDASlot(searchPane, new string[] { itemToSearch }, returnAll, runInfo, new DASlotSearchType[] { searchType }); }
        public byte[] GetDASlot(DAPane searchPane, string[] itemToSearch, bool returnAll, MacInfo runInfo, DASlotSearchType[] searchType) {
            byte paneSize = searchPane == DAPane.a ? (byte)61 : (byte)37;
            List<byte> slotMatches = new List<byte>();

            for(byte i = 1; i < paneSize; i++) {
                string slotInformation = GetDASlotInformation(searchPane, i, runInfo);
                for(byte ii = 0; ii < searchType.Length; ii++) {
                    if(searchType[ii] == DASlotSearchType.Contains) { if(!slotInformation.Contains(itemToSearch[ii])) break; } else if(searchType[ii] == DASlotSearchType.StartsWith) { if(!slotInformation.StartsWith(itemToSearch[ii], StringComparison.Ordinal)) break; } else if(searchType[ii] == DASlotSearchType.EndsWith) { if(!slotInformation.EndsWith(itemToSearch[ii], StringComparison.Ordinal)) break; } else if(searchType[ii] == DASlotSearchType.Exact) { if(!slotInformation.Equals(itemToSearch[ii], StringComparison.Ordinal)) break; }
                    if(ii == searchType.Length - 1)
                        if(returnAll) slotMatches.Add(i);
                        else return new byte[] { i };
                }
            }
            return slotMatches.ToArray();
        }

        public bool TryFindImageInDA(string searchImageFileName, IntPtr daWin, out Point imageCoordinates) {
            Point daCoords = new Point(0, 0);
            SafeNativeMethods.ClientToScreen(daWin, ref daCoords);

            if(!new FileInfo(searchImageFileName).Exists) { imageCoordinates = new Point(); return false; }
            using(Bitmap daImage = new Bitmap(640, 480), searchImage = new Bitmap(Image.FromFile(searchImageFileName))) {
                using(Graphics daWindowGraphic = Graphics.FromImage(daImage)) daWindowGraphic.CopyFromScreen(daCoords.X, daCoords.Y, 0, 0, new Size(640, 480));

                Color tmpColor;
                List<Color> searchCol = new List<Color>();
                List<short> searchColXOff = new List<short>(), searchColYOff = new List<short>();
                for(short siY = 0; siY < searchImage.Height; siY++)
                    for(short siX = 0; siX < searchImage.Width; siX++)
                        if((tmpColor = searchImage.GetPixel(siX, siY)).A != 0) {
                            searchCol.Add(tmpColor);
                            searchColXOff.Add(siX);
                            searchColYOff.Add(siY);
                        }
                short initX = 0, initY = 0;
                if(searchColXOff[0] != 0) { initX = searchColXOff[0]; for(short i = (short)(searchCol.Count - 1); i >= 0; i--) searchColXOff[i] = (short)(searchColXOff[i] - searchColXOff[0]); }
                if(searchColYOff[0] != 0) { initY = searchColYOff[0]; for(short i = (short)(searchCol.Count - 1); i >= 0; i--) searchColYOff[i] = (short)(searchColYOff[i] - searchColYOff[0]); }

                for(ushort daImageY = (ushort)searchImage.Height; daImageY < daImage.Height - searchImage.Height; daImageY++)
                    for(ushort daImageX = (ushort)searchImage.Width; daImageX < daImage.Width - searchImage.Width; daImageX++)
                        for(ushort i = 0; i < searchCol.Count; i++)
                            if(searchCol[i] != daImage.GetPixel(daImageX + searchColXOff[i], daImageY + searchColYOff[i])) break;
                            else if(i == searchCol.Count - 1) {
                                imageCoordinates = new Point(daImageX - initX + searchImage.Width / 2, daImageY - initY + searchImage.Height / 2);
                                return true;
                            }
            }
            imageCoordinates = new Point();
            return false;
        }

        private List<Thread> miscThreads = new List<Thread>(10);
        private readonly object miscThreadsLock = new object();
        //Plays Sound or Uses Attack in x amount of time (Can use "___.Join(Timeout.Infinate)" for the "waitForReturn"
        public Thread PlayDASoundIn(int timeToStartMillisecond, string fileName, MacInfo runInfo) { return PlayDASoundIn(timeToStartMillisecond, fileName, runInfo, 0); }
        public Thread PlayDASoundIn(int timeToStartMillisecond, string fileName, MacInfo runInfo, DAMacOps option) {
            Thread soundThread = new Thread(new ParameterizedThreadStart(PlayDASoundHelper));
            lock(miscThreadsLock) miscThreads.Add(soundThread);
            soundThread.Start(new object[] { timeToStartMillisecond, fileName, soundThread });

            if((option & DAMacOps.TIME_WAITFORRET) != 0) while(HKRunning[runInfo] && soundThread.IsAlive) Thread.Sleep(20);
            return soundThread;
        }
        private void PlayDASoundHelper(object runInfoO) {
            object[] myObjArr = runInfoO as object[];
            bool needsToPlay = true;

            try { Thread.Sleep((int)myObjArr[0]); } catch(ThreadInterruptedException) { needsToPlay = false; }
            if(needsToPlay && new FileInfo((string)myObjArr[1]).Exists) new System.Media.SoundPlayer((string)myObjArr[1]).Play();
            lock(miscThreadsLock) miscThreads.Remove((Thread)myObjArr[2]);
        }

        //This function returns when Macro is cancled or it used the attack
        public Thread UseDAAttackIn(int timeMillisecond, byte targetSlot, DAPane targetPane, bool mouseClick, MacInfo runInfo) { return UseDAAttackIn(timeMillisecond, targetSlot, targetPane, mouseClick, runInfo, null, null, 0); }
        public Thread UseDAAttackIn(int timeMillisecond, byte targetSlot, DAPane targetPane, bool mouseClick, MacInfo runInfo, DAMacOps option) { return UseDAAttackIn(timeMillisecond, targetSlot, targetPane, mouseClick, runInfo, null, null, option); }
        public Thread UseDAAttackIn(int timeMillisecond, byte targetSlot, DAPane targetPane, bool mouseClick, MacInfo runInfo, ushort? locX, ushort? locY) { return UseDAAttackIn(timeMillisecond, targetSlot, targetPane, mouseClick, runInfo, locX, locY, 0); }
        public Thread UseDAAttackIn(int timeMillisecond, byte targetSlot, DAPane targetPane, bool mouseClick, MacInfo runInfo, ushort? locX, ushort? locY, DAMacOps option) {
            Thread attackThread = new Thread(new ParameterizedThreadStart(UseDAAtkInHelper));
            lock(miscThreadsLock) miscThreads.Add(attackThread);
            attackThread.Start(new object[] { timeMillisecond, targetSlot, targetPane, mouseClick, runInfo, locX, locY, option, attackThread });

            if((option & DAMacOps.TIME_WAITFORRET) != 0) while(HKRunning[runInfo] && attackThread.IsAlive) Thread.Sleep(20);
            return attackThread;
        }
        private void UseDAAtkInHelper(object runInfoO) {
            object[] myObjArr = runInfoO as object[];
            bool needsToUse = true;

            try { Thread.Sleep((int)myObjArr[0]); } catch(ThreadInterruptedException) { needsToUse = false; }
            if(needsToUse) UseDAAttack((byte)myObjArr[1], (DAPane)myObjArr[2], (bool)myObjArr[3], (MacInfo)myObjArr[4], (ushort?)myObjArr[5], (ushort?)myObjArr[6], (DAMacOps)myObjArr[7]);
            lock(miscThreadsLock) miscThreads.Remove((Thread)myObjArr[8]);
        }

        public void ExitWindow(IntPtr handle) { lock(daMacLock) if(myProcesses.ContainsKey(handle) && !myProcesses[handle].CloseMainWindow()) myProcesses[handle].Kill(); }
        public void Base_ExitDAMac(MacInfo runInfo) { Application.Exit(); }
        public void Base_StopAllMacros(MacInfo runInfo) {
            new MyToolTipLib.MyToolTip("Pause(StopedMacros)", 2500, runInfo.Handle);
            StopMacros();
        }
        public void Base_ResetAllPointers(MacInfo runInfo) {
            new MyToolTipLib.MyToolTip("SPause(ResetedPointers)", 2500, runInfo.Handle);
            lock(daMacLock) {
                daGuiSkillPtr[runInfo.Handle] = null;
                SetDASkillPtr(runInfo);
            }
        }


        public ushort? AddHotkey(Keys hkKey, WindowModifiers hkMod, RunMacro hkThread, bool processCheck) {
            if(hkThread == null) return null;
            HKRunning.AddHK();

            hkKeys.Add(hkKey);
            hkMods.Add(hkMod);
            hkMacros.Add(hkThread);
            hkAtomNames.Add(SafeNativeMethods.GlobalAddAtom("hk" + hkAtomNames.Count.ToString(CultureInfo.InvariantCulture) + "_daMac"));
            hkCheckForDA.Add(processCheck);

            for(byte i = 0; i < hkKeys.Count; i++)
                if(hkKeys[i] == hkKey && hkMods[i] == hkMod && hkMacros[i] == hkThread) {
                    SafeNativeMethods.RegisterHotKey(mainHandle, hkAtomNames[i], hkMods[i], (uint)hkKeys[i]);
                    break;
                }
            return hkAtomNames[hkAtomNames.Count - 1];
        }
        public void RemoveHotkey(Keys hkKey, WindowModifiers hkMod) {
            for(byte i = 0; i < hkKeys.Count; i++)
                if(hkKeys[i] == hkKey && hkMods[i] == hkMod) {
                    RemoveHotkey_Helper(i);
                    break;
                }
        }
        public void RemoveHotkey(ushort hkAtom) {
            for(byte i = 0; i < hkAtomNames.Count; i++)
                if(hkAtomNames[i] == hkAtom) {
                    RemoveHotkey_Helper(i);
                    break;
                }
        }
        private void RemoveHotkey_Helper(byte hkIndex) {
            HKRunning.RemoveHK(hkIndex);
            hkKeys.RemoveAt(hkIndex);
            hkMods.RemoveAt(hkIndex);
            hkMacros.RemoveAt(hkIndex);
            hkCheckForDA.RemoveAt(hkIndex);

            SafeNativeMethods.UnregisterHotKey(mainHandle, hkAtomNames[hkIndex]);
            SafeNativeMethods.GlobalDeleteAtom(hkAtomNames[hkIndex]);
            hkAtomNames.RemoveAt(hkIndex);
        }

        public byte? GetMacNumber(ushort hkAtom) { for(byte i = 0; i < hkAtomNames.Count; i++) if(hkAtomNames[i] == hkAtom) return i; return null; }
        public byte? GetMacNumber(Keys hkKey, WindowModifiers hkMod) { for(byte i = 0; i < hkKeys.Count; i++) if(hkKeys[i] == hkKey && hkMods[i] == hkMod) return i; return null; }
        public byte? GetMacNumber(string methodName) { for(byte i = 0; i < hkMacros.Count; i++) if(hkMacros[i].Method.Name == methodName) return i; return null; }

        public bool CheckHKOn(ushort hkAtom, IntPtr daWin) {
            byte? macNum = GetMacNumber(hkAtom);
            return !macNum.HasValue ? false : HKRunning[new MacInfo(daWin, macNum.Value)];
        }
        public bool CheckHKOn(Keys hkKey, WindowModifiers hkMod, IntPtr daWin) {
            byte? macNum = GetMacNumber(hkKey, hkMod);
            return !macNum.HasValue ? false : HKRunning[new MacInfo(daWin, macNum.Value)];
        }
        public bool CheckHKOn(string methodName, IntPtr daWin) {
            byte? macNum = GetMacNumber(methodName);
            return !macNum.HasValue ? false : HKRunning[new MacInfo(daWin, macNum.Value)];
        }

        public bool StartHK(ushort hkAtom) {
            byte? i = GetMacNumber(hkAtom);
            return !i.HasValue ? false : StartHK(i.Value);
        }
        public bool StartHK(Keys hkKey, WindowModifiers hkMod) {
            byte? i = GetMacNumber(hkKey, hkMod);
            return !i.HasValue ? false : StartHK(i.Value);
        }
        public bool StartHK(string methodName) {
            byte? i = GetMacNumber(methodName);
            return !i.HasValue ? false : StartHK(i.Value);
        }
        public bool StartHK(byte macNumber) {
            MacInfo sendInfo = new MacInfo(SafeNativeMethods.GetForegroundWindow(), macNumber);

            if(HKRunning.IsThreadInMacro(sendInfo)) HKRunning[sendInfo] = false;
            else if(!hkCheckForDA[macNumber] || StartProcessCheck(sendInfo.Handle)) ThreadPool.QueueUserWorkItem(new WaitCallback(StartHKThread), sendInfo);
            else {
                //Mimic the Keypress since DA Isn't Up
                List<Keys> modsSends = new List<Keys>();
                if((hkMods[macNumber] & WindowModifiers.MOD_SHIFT) != 0) modsSends.Add(Keys.ShiftKey);
                if((hkMods[macNumber] & WindowModifiers.MOD_CONTROL) != 0) modsSends.Add(Keys.ControlKey);
                if((hkMods[macNumber] & WindowModifiers.MOD_ALT) != 0) modsSends.Add(Keys.Menu);

                if(modsSends.Count == 0) SendKeyPM(hkKeys[macNumber], null, sendInfo.Handle, DAMacKeyOps.SK_REAL);
                else SendKeyPM(hkKeys[macNumber], modsSends.ToArray(), sendInfo.Handle, DAMacKeyOps.SK_REAL);

                return false;
            }
            return true;
        }
        private void StartHKThread(object runInfoO) {
            MacInfo runInfo = runInfoO as MacInfo;
            HKRunning[runInfo] = true;
            hkMacros[runInfo.MacNumber](runInfo);
            HKRunning.ThreadExitingMacro(runInfo);
        }

        //Throws IOExceptions, Parcing Exceptions, NullArgumentExceptions(If daDataFile Format is Incorrect)
        private Dictionary<string, Pointer> daPtrs = new Dictionary<string, Pointer>(17);
        private void SetDAPointerInfo(string daDataFile) {
            foreach(XElement A in XElement.Load(daDataFile).Element("Pointers").Elements()) {
                Pointer newPointer = new Pointer() { BasePtr = UInt32.Parse(A.Value, NumberStyles.HexNumber, CultureInfo.InvariantCulture) };
                foreach(XAttribute B in A.Attributes().OrderBy(x => x.Name.LocalName))
                    if(Regex.Match(B.Name.LocalName, @"^o(\d)+$").Success) newPointer.Offs.Add(UInt16.Parse(B.Value, NumberStyles.HexNumber, CultureInfo.InvariantCulture));
                    else newPointer.Extras.Add(B.Name.LocalName, UInt16.Parse(B.Value, NumberStyles.HexNumber, CultureInfo.InvariantCulture));
                daPtrs.Add(A.Name.LocalName, newPointer);
            }
        }
        private void StopMacros() {
            HKRunning.SetToFalse();
            lock(miscThreadsLock) foreach(Thread i in miscThreads) i.Interrupt();
        }


        private bool disposed;
        public void Dispose() { Dispose(true); GC.SuppressFinalize(this); }
        protected virtual void Dispose(bool disposing) {
            if(!disposed) {
                if(disposing) {
                    StopMacros();
                    while(hkKeys.Count > 0) RemoveHotkey(hkKeys[0], hkMods[0]);
                    foreach(KeyValuePair<IntPtr, MyMemoryReadSearchLib.MyMemoryReadSearch> i in pReader) i.Value.Dispose();
                    foreach(Process A in myProcesses.Values) A.Dispose();
                }
                disposed = true;
            }
        }


        private MyDAMapLib.MyDAMap myDAMap;
        private MyDAMapLib.MyDAMap MyDAMap { get { if(myDAMap == null) myDAMap = new MyDAMapLib.MyDAMap(); return myDAMap; } }
        public bool MoveTo(byte locX, byte locY, MacInfo runInfo) { return MoveTo(locX, locY, runInfo, 0, false); }
        public bool MoveTo(byte locX, byte locY, MacInfo runInfo, bool doorWalk) { return MoveTo(locX, locY, runInfo, 0, doorWalk); }
        public bool MoveTo(byte locX, byte locY, MacInfo runInfo, byte range) { return MoveTo(locX, locY, runInfo, range, false); }
        private bool MoveTo(byte locX, byte locY, MacInfo runInfo, byte range, bool doorWalk) {
            //Get the Current Map# and Build mapFolder String
            ushort? mapNum = GetDAMapNumber(runInfo);
            if(!mapNum.HasValue) { Thread.Sleep(550); if(!(mapNum = GetDAMapNumber(runInfo)).HasValue) return false; }
            string mapFolder = (File.Exists("DAData.xml") ? XElement.Load("DAData.xml").Element("Data").Element("MapFolder").Value : @"C:\Program Files\KRU\Dark Ages\maps\") + "lod" + mapNum.Value.ToString(CultureInfo.InvariantCulture) + ".map";

            //Get Map Information Like Axis I'm Gonna Use and Walls
            List<ushort[]> mapFactors = MyDAMapLib.MyDAMap.GetFactors(mapFolder);
            if(mapFactors == null) return false; //if(MapFactors == null) { Thread.Sleep(550); if((MapFactors = currMap.GetFactors(mapFolder)) == null) return false; }

            ushort[] mapMaxs = mapFactors[(byte)(mapFactors.Count / 2)];
            char[,] currMapWalls = MyDAMap.GetMakeDAMapChar(mapFolder, mapMaxs[0], mapMaxs[1]);
            if(currMapWalls == null) return false; //if(currMapWalls == null) { Thread.Sleep(550); if((currMapWalls = currMap.GetMakeDAMapChar(mapFolder, mapMaxs[0], mapMaxs[1])) == null) return false; }

        ReGetPath:
            //Init the Variables and Enqueue the first element
            MyPriorityQueue searchQueue = new MyPriorityQueue();
            byte[] cc = GetDACharCoordinates(runInfo);
            if(cc == null) return false; //if(cc == null) { Thread.Sleep(550); if((cc = GetDACoords(runInfo)) == null) return false; }
            for(byte i = 0; i < mapMaxs[1]; i++) for(byte j = 0; j < mapMaxs[0]; j++) if(currMapWalls[j, i] == 'S' || currMapWalls[j, i] == 'R' || currMapWalls[j, i] == 'D' || currMapWalls[j, i] == 'L' || currMapWalls[j, i] == 'U') currMapWalls[j, i] = (char)0;
            currMapWalls[cc[0], cc[1]] = 'S';

            while(HKRunning[runInfo] && (cc[0] != locX || cc[1] != locY)) {
                //Add 4 Blocks to the Queue
                if(cc[0] > 0 && currMapWalls[cc[0] - 1, cc[1]] == (char)0) {
                    currMapWalls[cc[0] - 1, cc[1]] = 'R';
                    searchQueue.Enqueue(new byte[] { (byte)(cc[0] - 1), cc[1] }, GetManhattanDis((byte)(cc[0] - 1), cc[1], locX, locY));
                }
                if(cc[1] > 0 && currMapWalls[cc[0], cc[1] - 1] == (char)0) {
                    currMapWalls[cc[0], cc[1] - 1] = 'D';
                    searchQueue.Enqueue(new byte[] { cc[0], (byte)(cc[1] - 1) }, GetManhattanDis(cc[0], (byte)(cc[1] - 1), locX, locY));
                }
                if(cc[0] < mapMaxs[0] - 1 && currMapWalls[cc[0] + 1, cc[1]] == (char)0) {
                    currMapWalls[cc[0] + 1, cc[1]] = 'L';
                    searchQueue.Enqueue(new byte[] { (byte)(cc[0] + 1), cc[1] }, GetManhattanDis((byte)(cc[0] + 1), cc[1], locX, locY));
                }
                if(cc[1] < mapMaxs[1] - 1 && currMapWalls[cc[0], cc[1] + 1] == (char)0) {
                    currMapWalls[cc[0], cc[1] + 1] = 'U';
                    searchQueue.Enqueue(new byte[] { cc[0], (byte)(cc[1] + 1) }, GetManhattanDis(cc[0], (byte)(cc[1] + 1), locX, locY));
                }
                if(searchQueue.isEmpty) {
                    bool ending = true;
                    for(byte i = 0; i < mapMaxs[1]; i++)
                        for(byte j = 0; j < mapMaxs[0]; j++)
                            if(currMapWalls[j, i] == 'A') {
                                ending = false;
                                currMapWalls[j, i] = (char)0;
                            }
                    if(ending) return false;
                    else goto ReGetPath;
                } else cc = searchQueue.Dequeue();
            }

            //Build Path Array from Gold (cc[0] / cc[1] should Equal dX/dY at this point)
            List<char> path = new List<char>();
            while(HKRunning[runInfo] && currMapWalls[cc[0], cc[1]] != 'S')
                switch(currMapWalls[cc[0], cc[1]]) {
                    case 'R': path.Add('L'); cc[0]++; break;
                    case 'D': path.Add('U'); cc[1]++; break;
                    case 'L': path.Add('R'); cc[0]--; break;
                    case 'U': path.Add('D'); cc[1]--; break;
                    default: return false;
                }
            path.Reverse();


            //Move To
            byte[] ccChk;
            char prevDir = (char)0;
            byte wallChkLeft = 4;
            foreach(char A in path) {
                if(!HKRunning[runInfo]) return false;
                else if(GetManhattanDis(cc[0], cc[1], locX, locY) <= range) break;
                switch(A) {
                    case 'L': cc[0]--;
                        while(HKRunning[runInfo] && (ccChk = GetDACharCoordinates(runInfo)) != null && cc[0] != ccChk[0]) {
                            if(--wallChkLeft == 0)
                                if(mapNum != GetDAMapNumber(runInfo)) return doorWalk ? true : false; //false should be "Try to Go Back"
                                else { if(currMapWalls[cc[0], cc[1]] != 'W' || currMapWalls[cc[0], cc[1]] != 'A') currMapWalls[cc[0], cc[1]] = 'A'; goto ReGetPath; }
                            if(prevDir != 'L') SendKeyPM(Keys.Left, runInfo.Handle);
                            SendKeyPM(Keys.Left, runInfo.Handle);
                            Thread.Sleep(550);
                        } wallChkLeft = 4; break;
                    case 'U': cc[1]--;
                        while(HKRunning[runInfo] && (ccChk = GetDACharCoordinates(runInfo)) != null && cc[1] != ccChk[1]) {
                            if(--wallChkLeft == 0)
                                if(mapNum != GetDAMapNumber(runInfo)) return doorWalk ? true : false; //false should be "Try to Go Back"
                                else { if(currMapWalls[cc[0], cc[1]] != 'W' || currMapWalls[cc[0], cc[1]] != 'A') currMapWalls[cc[0], cc[1]] = 'A'; goto ReGetPath; }
                            if(prevDir != 'U') SendKeyPM(Keys.Up, runInfo.Handle);
                            SendKeyPM(Keys.Up, runInfo.Handle);
                            Thread.Sleep(550);
                        } wallChkLeft = 4; break;
                    case 'R': cc[0]++;
                        while(HKRunning[runInfo] && (ccChk = GetDACharCoordinates(runInfo)) != null && cc[0] != ccChk[0]) {
                            if(--wallChkLeft == 0)
                                if(mapNum != GetDAMapNumber(runInfo)) return doorWalk ? true : false; //false should be "Try to Go Back"
                                else { if(currMapWalls[cc[0], cc[1]] != 'W' || currMapWalls[cc[0], cc[1]] != 'A') currMapWalls[cc[0], cc[1]] = 'A'; goto ReGetPath; }
                            if(prevDir != 'R') SendKeyPM(Keys.Right, runInfo.Handle);
                            SendKeyPM(Keys.Right, runInfo.Handle);
                            Thread.Sleep(550);
                        } wallChkLeft = 4; break;
                    case 'D': cc[1]++;
                        while(HKRunning[runInfo] && (ccChk = GetDACharCoordinates(runInfo)) != null && cc[1] != ccChk[1]) {
                            if(--wallChkLeft == 0)
                                if(mapNum != GetDAMapNumber(runInfo)) return doorWalk ? true : false; //false should be "Try to Go Back"
                                else { if(currMapWalls[cc[0], cc[1]] != 'W' || currMapWalls[cc[0], cc[1]] != 'A') currMapWalls[cc[0], cc[1]] = 'A'; goto ReGetPath; }
                            if(prevDir != 'D') SendKeyPM(Keys.Down, runInfo.Handle);
                            SendKeyPM(Keys.Down, runInfo.Handle);
                            Thread.Sleep(550);
                        } wallChkLeft = 4; break;
                } prevDir = A;
            }
            SendKeyPM(Keys.F5, runInfo.Handle); Thread.Sleep(1025);
            if(mapNum != GetDAMapNumber(runInfo)) return doorWalk ? true : false; //false should be "Try to Go Back"
            else if((cc = GetDACharCoordinates(runInfo)) != null && GetManhattanDis(cc[0], cc[1], locX, locY) <= range) return true;
            else goto ReGetPath;
        }
        private static byte GetManhattanDis(byte cX, byte cY, byte gX, byte gY) { return (byte)((cX > gX ? cX - gX : gX - cX) + (cY > gY ? cY - gY : gY - cY)); }
        private class MyPriorityQueue {
            private List<KeyValuePair<byte, byte[]>> data;
            public bool isEmpty;

            public MyPriorityQueue() {
                data = new List<KeyValuePair<byte, byte[]>>();
                isEmpty = true;
            }
            public void Enqueue(byte[] item, byte priority) {
                data.Add(new KeyValuePair<byte, byte[]>(priority, item));
                data.Sort((x, y) => x.Key < y.Key ? 1 : (x.Key > y.Key ? -1 : 0));
                if(isEmpty) isEmpty = false;
            }
            public byte[] Dequeue() {
                byte[] rtn = data[data.Count - 1].Value;
                data.RemoveAt(data.Count - 1);
                if(data.Count == 0) isEmpty = true;
                return rtn;
            }
        }
    }
}