using System;
using System.Globalization;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using System.Xml.Linq;

namespace DASearcher {
    /// <summary>
    /// Summary description for Searcher.
    /// </summary>
    public class Searcher {
        public MyMemoryReadSearchLib.MyMemoryReadSearch pReader = new MyMemoryReadSearchLib.MyMemoryReadSearch();
        private byte[] memory = null;

        private IntPtr DarkageshWnd = IntPtr.Zero;

        //Change to Use DAMac's GetPointer In Future
        private int InfoBasePtr = Int32.Parse(XElement.Load("DAData.xml").Element("Pointers").Element("MapNum").Value, NumberStyles.HexNumber, CultureInfo.InvariantCulture);

        private int YAddress = 0;
        private int YAddressOff = Int32.Parse(XElement.Load("DAData.xml").Element("Pointers").Element("Coords").Attribute("o1").Value, NumberStyles.HexNumber, CultureInfo.InvariantCulture);
        private int XAddress = 0;
        private int XAddressOff = Int32.Parse(XElement.Load("DAData.xml").Element("Pointers").Element("Coords").Attribute("o1").Value, NumberStyles.HexNumber, CultureInfo.InvariantCulture) + Int32.Parse(XElement.Load("DAData.xml").Element("Pointers").Element("Coords").Attribute("size").Value, NumberStyles.HexNumber, CultureInfo.InvariantCulture);
        private int MapNumAddress = 0;
        private int MapNumAddressOff = Int32.Parse(XElement.Load("DAData.xml").Element("Pointers").Element("MapNum").Attribute("o1").Value, NumberStyles.HexNumber, CultureInfo.InvariantCulture);

        public int X, Y, MaxX, MaxY, MapNum, MapNumStart = 0;
        public int delay = 550;
        public int WallChecks = 3;
        private string PrevDirection = "";
        public bool Running = false;
        public char[,] mainMap;
        private MainFrame MainParent;

        //Know if Map is Loaded
        public int mapLoaded = 0;


        public Searcher(MainFrame MainF) {
            MainParent = MainF;
            MakeMainMap(128, 128);
        }

        public void StartSearch() {
            Running = StartProcessCheck();

            memory = pReader.ReadProcessMemory((IntPtr)MapNumAddress, new UIntPtr(2));
            if(memory != null) MapNumStart = (memory[0] + 256 * memory[1]);
            else if(MainParent.StopBtn.Enabled) MainParent.StopBtn_Click(null, null);

            Thread.Sleep(500);

            while(Running) {
                byte CatchExceptions = 10;
                try {
                    UpdateValues();
                    if(mainMap[X, Y] == (char)0) mainMap[X, Y] = 'C';
                    FindAndMoveToClosest();
                    CatchExceptions = 10;
                } catch(Exception) { Thread.Sleep(250); if(--CatchExceptions == 0) Running = false; }
            }
        }

        public bool StartProcessCheck() {
            System.Diagnostics.Process[] myProcesses = System.Diagnostics.Process.GetProcessesByName("Darkages");

            if(myProcesses.Length == 0) {
                if(MainParent.StopBtn.Enabled) MainParent.StopBtn_Click(null, null);
                MessageBox.Show("No DarkAges Process found!");
                return false;
            }

            pReader.OpenProcess((uint)myProcesses[0].Id);

            DarkageshWnd = myProcesses[0].MainWindowHandle;

            memory = pReader.ReadProcessMemory((IntPtr)InfoBasePtr, new UIntPtr(4));
            int refTemp = (memory[0] + 256 * memory[1] + 256 * 256 * memory[2] + 256 * 256 * 256 * memory[3]);

            XAddress = refTemp + XAddressOff;
            YAddress = refTemp + YAddressOff;
            MapNumAddress = refTemp + MapNumAddressOff;

            return true;
        }

        private void FindAndMoveToClosest() {
            bool FoundEmptySpot = false;
            char[,] mainMapCheck = new char[MaxX, MaxY];
            System.Collections.Generic.Queue<int[]> SearchQue = new System.Collections.Generic.Queue<int[]>();
            int CurX = X;
            int CurY = Y;


            mainMapCheck[CurX, CurY] = 'S';
            while(!FoundEmptySpot) {
                //Check Left
                if(CurX > 0 && mainMap[CurX - 1, CurY] == 0) {
                    mainMapCheck[CurX - 1, CurY] = 'R';
                    CurX--;
                    FoundEmptySpot = true;
                } else {
                    if(CurX > 0 && (mainMap[CurX - 1, CurY] == 'C' || mainMap[CurX - 1, CurY] == 'D') && mainMapCheck[CurX - 1, CurY] == 0) {
                        mainMapCheck[CurX - 1, CurY] = 'R';
                        SearchQue.Enqueue(new int[] { CurX - 1, CurY });
                    }

                    //Check Up
                    if(CurY > 0 && mainMap[CurX, CurY - 1] == 0) {
                        mainMapCheck[CurX, CurY - 1] = 'D';
                        CurY--;
                        FoundEmptySpot = true;
                    } else {
                        if(CurY > 0 && (mainMap[CurX, CurY - 1] == 'C' || mainMap[CurX, CurY - 1] == 'D') && mainMapCheck[CurX, CurY - 1] == 0) {
                            mainMapCheck[CurX, CurY - 1] = 'D';
                            SearchQue.Enqueue(new int[] { CurX, CurY - 1 });
                        }

                        //Check Right
                        if(CurX < MaxX - 1 && mainMap[CurX + 1, CurY] == 0) {
                            mainMapCheck[CurX + 1, CurY] = 'L';
                            CurX++;
                            FoundEmptySpot = true;
                        } else {
                            if(CurX < MaxX - 1 && (mainMap[CurX + 1, CurY] == 'C' || mainMap[CurX + 1, CurY] == 'D') && mainMapCheck[CurX + 1, CurY] == 0) {
                                mainMapCheck[CurX + 1, CurY] = 'L';
                                SearchQue.Enqueue(new int[] { CurX + 1, CurY });
                            }

                            //Check Down
                            if(CurY < MaxY - 1 && mainMap[CurX, CurY + 1] == 0) {
                                mainMapCheck[CurX, CurY + 1] = 'U';
                                CurY++;
                                FoundEmptySpot = true;
                            } else {
                                if(CurY < MaxY - 1 && (mainMap[CurX, CurY + 1] == 'C' || mainMap[CurX, CurY + 1] == 'D') && mainMapCheck[CurX, CurY + 1] == 0) {
                                    mainMapCheck[CurX, CurY + 1] = 'U';
                                    SearchQue.Enqueue(new int[] { CurX, CurY + 1 });
                                }

                                //Check if anything else to check & Update CurX and CurY
                                if(SearchQue.Count == 0) {
                                    if(MainParent.LBLAutoClear.Checked) {
                                        ClearSearched();
                                        return;
                                    } else if(MainParent.StopBtn.Enabled) MainParent.StopBtn_Click(null, null);
                                } else {
                                    int[] TempCoords = new int[2];
                                    TempCoords = SearchQue.Dequeue();
                                    CurX = TempCoords[0];
                                    CurY = TempCoords[1];
                                }
                            }
                        }
                    }
                }
            }
            MoveTo(CurX, CurY, mainMapCheck);
        }

        private void MoveTo(int NewX, int NewY, char[,] BPath) {
            char[] Path = new char[(MaxX * MaxY) / 2 + (MaxX + MaxY) / 2];
            int loc = 0;
            int XCur = NewX;
            int YCur = NewY;
            int counter = 1;


            //Build Path Array from BPath
            while(BPath[XCur, YCur] != 'S') {
                if(BPath[XCur, YCur] == 'R') {
                    Path[loc++] = 'L';
                    XCur++;
                } else if(BPath[XCur, YCur] == 'D') {
                    Path[loc++] = 'U';
                    YCur++;
                } else if(BPath[XCur, YCur] == 'L') {
                    Path[loc++] = 'R';
                    XCur--;
                } else if(BPath[XCur, YCur] == 'U') {
                    Path[loc++] = 'D';
                    YCur--;
                }
            }

            //Process the 'C' Moves
            while(--loc > 0) {
                UpdateValues();
                counter = 1;
                XCur = X;
                YCur = Y;

                if(Path[loc] == 'L') {
                    while(XCur == X) {
                        if(++counter > WallChecks) {
                            if(mainMap[X - 1, Y] != 'W') if(mainMap[X - 1, Y] != 'D') mainMap[X - 1, Y] = 'A'; else mainMap[X - 1, Y] = 'W';
                            return;
                        }
                        MoveLeft();
                        Thread.Sleep(delay);
                        UpdateValues();
                    }
                } else if(Path[loc] == 'U') {
                    while(YCur == Y) {
                        if(++counter > WallChecks) {
                            if(mainMap[X, Y - 1] != 'W') if(mainMap[X, Y - 1] != 'D') mainMap[X, Y - 1] = 'A'; else mainMap[X, Y - 1] = 'W';
                            return;
                        }
                        MoveUp();
                        Thread.Sleep(delay);
                        UpdateValues();
                    }
                } else if(Path[loc] == 'R') {
                    while(XCur == X) {
                        if(++counter > WallChecks) {
                            if(mainMap[X + 1, Y] != 'W') if(mainMap[X + 1, Y] != 'D') mainMap[X + 1, Y] = 'A'; else mainMap[X + 1, Y] = 'W';
                            return;
                        }
                        MoveRight();
                        Thread.Sleep(delay);
                        UpdateValues();
                    }
                } else if(Path[loc] == 'D') {
                    while(YCur == Y) {
                        if(++counter > WallChecks) {
                            if(mainMap[X, Y + 1] != 'W') if(mainMap[X, Y + 1] != 'D') mainMap[X, Y + 1] = 'A'; else mainMap[X, Y + 1] = 'W';
                            return;
                        }
                        MoveDown();
                        Thread.Sleep(delay);
                        UpdateValues();
                    }
                }
            }

            // Process's the Last Move
            counter = 1;
            XCur = X;
            YCur = Y;

            if(Path[loc] == 'L') {
                while(XCur == X) {
                    if(++counter > WallChecks) {
                        if(mainMap[X - 1, Y] != 'W') if(mainMap[X - 1, Y] != 'D') mainMap[X - 1, Y] = 'A'; else mainMap[X - 1, Y] = 'W';
                        break;
                    }
                    MoveLeft();
                    Thread.Sleep(delay);
                    UpdateValues();
                }
                if(MapNumStart != MapNum) {
                    if(mainMap[XCur - 1, YCur] != 'W') if(mainMap[XCur - 1, YCur] != 'D') mainMap[XCur - 1, YCur] = 'A'; else mainMap[XCur - 1, YCur] = 'W';
                    ThroughDoorMoving("Left");
                }
            } else if(Path[loc] == 'U') {
                while(YCur == Y) {
                    if(++counter > WallChecks) {
                        if(mainMap[X, Y - 1] != 'W') if(mainMap[X, Y - 1] != 'D') mainMap[X, Y - 1] = 'A'; else mainMap[X, Y - 1] = 'W';
                        break;
                    }
                    MoveUp();
                    Thread.Sleep(delay);
                    UpdateValues();
                }
                if(MapNumStart != MapNum) {
                    if(mainMap[XCur, YCur - 1] != 'W') if(mainMap[XCur, YCur - 1] != 'D') mainMap[XCur, YCur - 1] = 'A'; else mainMap[XCur, YCur - 1] = 'W';
                    ThroughDoorMoving("Up");
                }
            } else if(Path[loc] == 'R') {
                while(XCur == X) {
                    if(++counter > WallChecks) {
                        if(mainMap[X + 1, Y] != 'W') if(mainMap[X + 1, Y] != 'D') mainMap[X + 1, Y] = 'A'; else mainMap[X + 1, Y] = 'W';
                        break;
                    }
                    MoveRight();
                    Thread.Sleep(delay);
                    UpdateValues();
                }
                if(MapNumStart != MapNum) {
                    if(mainMap[XCur + 1, YCur] != 'W') if(mainMap[XCur + 1, YCur] != 'D') mainMap[XCur + 1, YCur] = 'A'; else mainMap[XCur + 1, YCur] = 'W';
                    ThroughDoorMoving("Right");
                }
            } else if(Path[loc] == 'D') {
                while(YCur == Y) {
                    if(++counter > WallChecks) {
                        if(mainMap[X, Y + 1] != 'W') if(mainMap[X, Y + 1] != 'D') mainMap[X, Y + 1] = 'A'; else mainMap[X, Y + 1] = 'W';
                        break;
                    }
                    MoveDown();
                    Thread.Sleep(delay);
                    UpdateValues();
                }
                if(MapNumStart != MapNum) {
                    if(mainMap[XCur, YCur + 1] != 'W') if(mainMap[XCur, YCur + 1] != 'D') mainMap[XCur, YCur + 1] = 'A'; else mainMap[XCur, YCur + 1] = 'W';
                    ThroughDoorMoving("Down");
                }
            }
        }

        private void ThroughDoorMoving(string direction) {
            int counter = 0;

            while(MapNumStart != MapNum) {
                if(counter >= 3 && counter < 5) {
                    if(direction == "Left") MoveDown();
                    else if(direction == "Up") MoveLeft();
                    else if(direction == "Right") MoveUp();
                    else if(direction == "Down") MoveRight();
                    Thread.Sleep(1050);
                    UpdateValues();
                } else if(counter >= 5 && counter < 9) {
                    if(direction == "Left") MoveUp();
                    else if(direction == "Up") MoveRight();
                    else if(direction == "Right") MoveDown();
                    else if(direction == "Down") MoveLeft();
                    Thread.Sleep(1050);
                    UpdateValues();
                } else if(counter >= 9) {
                    if(MainParent.StopBtn.Enabled) MainParent.StopBtn_Click(null, null);
                } else {
                    if(direction == "Left") MoveRight();
                    else if(direction == "Up") MoveDown();
                    else if(direction == "Right") MoveLeft();
                    else if(direction == "Down") MoveUp();
                    Thread.Sleep(1050);
                    UpdateValues();
                }
                counter++;
            }
        }

        // KeyDown	lParam = (0/4)1(50/4B/48/4D)0001   (4 if Continues, 0 for first)
        // KeyUp	lParam = C1(50/4B/48/4D)0001
        private void MoveLeft() {
            if(PrevDirection == "Left") {
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Left, (UIntPtr)0x014B0001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Left, (UIntPtr)0xC14B0001);
            } else {
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Left, (UIntPtr)0x014B0001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Left, (UIntPtr)0xC14B0001);
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Left, (UIntPtr)0x014B0001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Left, (UIntPtr)0xC14B0001);
                PrevDirection = "Left";
            }
        }
        private void MoveUp() {
            if(PrevDirection == "Up") {
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Up, (UIntPtr)0x01480001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Up, (UIntPtr)0xC1480001);
            } else {
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Up, (UIntPtr)0x01480001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Up, (UIntPtr)0xC1480001);
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Up, (UIntPtr)0x01480001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Up, (UIntPtr)0xC1480001);
                PrevDirection = "Up";
            }
        }
        private void MoveRight() {
            if(PrevDirection == "Right") {
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Right, (UIntPtr)0x014D0001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Right, (UIntPtr)0xC14D0001);
            } else {
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Right, (UIntPtr)0x014D0001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Right, (UIntPtr)0xC14D0001);
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Right, (UIntPtr)0x014D0001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Right, (UIntPtr)0xC14D0001);
                PrevDirection = "Right";
            }
        }
        private void MoveDown() {
            if(PrevDirection == "Down") {
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Down, (UIntPtr)0x01500001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Down, (UIntPtr)0xC1500001);
            } else {
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Down, (UIntPtr)0x01500001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Down, (UIntPtr)0xC1500001);
                MainFrame.PostMessage(DarkageshWnd, 0x0100, (UIntPtr)Keys.Down, (UIntPtr)0x01500001);
                MainFrame.PostMessage(DarkageshWnd, 0x0101, (UIntPtr)Keys.Down, (UIntPtr)0xC1500001);
                PrevDirection = "Down";
            }
        }

        private void MakeMainMap(int GivenX, int GivenY) {
            MaxX = GivenX;
            MaxY = GivenY;

            mainMap = new char[MaxX, MaxY];
        }

        public void UpdateValues() {
            if((memory = pReader.ReadProcessMemory((IntPtr)XAddress, (UIntPtr)1)) == null) { if(MainParent.StopBtn.Enabled) MainParent.StopBtn_Click(null, null); } else X = (memory[0]);
            if((memory = pReader.ReadProcessMemory((IntPtr)YAddress, (UIntPtr)1)) == null) { if(MainParent.StopBtn.Enabled) MainParent.StopBtn_Click(null, null); } else Y = (memory[0]);
            if((memory = pReader.ReadProcessMemory((IntPtr)MapNumAddress, (UIntPtr)2)) == null) { if(MainParent.StopBtn.Enabled) MainParent.StopBtn_Click(null, null); } else MapNum = (memory[0] + 256 * memory[1]);
        }

        public void SaveMap(string name) {
            StreamWriter SaveFile;
            try { SaveFile = File.CreateText(name); } catch(Exception e) {
                MessageBox.Show("Couldn't Create Save File" + e.Message);
                return;
            }

            SaveFile.Write(MaxX.ToString() + "\n" + MaxY.ToString() + "\n" + mapLoaded.ToString() + "\n");
            for(int i = 0; i < MaxY; i++) {
                for(int c = 0; c < MaxX; c++) {
                    SaveFile.Write(mainMap[c, i]);
                }
            }
            SaveFile.Close();
        }

        public void LoadSavedMap(string name) {
            string[] SavedFile;

            try {
                using(StreamReader OpenFile = File.OpenText(name)) SavedFile = OpenFile.ReadToEnd().Split('\n');
                if(SavedFile.Length != 4) {
                    MessageBox.Show("Coudln't Open File - Wasn't a DASearcher Saved File");
                    return;
                }
            } catch(Exception e) {
                MessageBox.Show("Coudln't Open File" + e.Message);
                return;
            }

            int c = 0;
            int i = 0;

            try { MakeMainMap(Convert.ToInt16(SavedFile[0]), Convert.ToInt16(SavedFile[1])); mapLoaded = Convert.ToInt16(SavedFile[2]); } catch(Exception e) {
                MessageBox.Show("Coudln't Open Map File: " + e.Message);
                return;
            }

            for(int j = 0; j < SavedFile[3].Length; j++) {
                if(c >= MaxX || i >= MaxY) {
                    MessageBox.Show("Files in Wrong Format");
                    return;
                }
                mainMap[c++, i] = SavedFile[3][j];
                if(c == MaxX) {
                    i++;
                    c = 0;
                }
            }
        }

        public void ConvertCurrMap(int ConvertedX, int ConvertedY) {
            char[,] NewMap = new char[ConvertedX, ConvertedY];

            for(int i = 0; i < ConvertedX; i++) {
                for(int c = 0; c < ConvertedY; c++) {
                    if(c >= MaxX || i >= MaxY) NewMap[c, i] = (char)0;
                    else NewMap[c, i] = mainMap[c, i];
                }
            }

            MaxX = ConvertedX;
            MaxY = ConvertedY;
            mainMap = NewMap;
        }

        public void ClearMap() {
            mainMap = new char[MaxX, MaxY];
        }

        public void ClearSearched() {
            for(int i = 0; i < MaxY; i++) {
                for(int c = 0; c < MaxX; c++) {
                    if(mainMap[c, i] == 'C' || mainMap[c, i] == 'A') mainMap[c, i] = (char)0;
                }
            }
        }

        public void PathSearcher() {
            Running = StartProcessCheck();

            if(!MainParent.AutoLoad.Checked) ClearMap();
            while(Running) {
                UpdateValues();
                mainMap[X, Y] = 'P';
                Thread.Sleep(250);
            }
        }

        public void PathSearcherConvert() {
            for(int i = 0; i < MaxY; i++) {
                for(int c = 0; c < MaxX; c++) {
                    if(mainMap[c, i] == 'P') mainMap[c, i] = (char)0;
                    else if(mainMap[c, i] != 'W') mainMap[c, i] = 'D';
                }
            }
        }
    }
}